/*
 * Copyright 2017 NXP
 */

#ifndef __GPUPERFCNT_H
#define __GPUPERFCNT_H

/**
 * \brief Library for retrieving various GPU information.
 */

#ifdef __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>

#define STR(x)	#x
#define XSTR(x)	STR(x)

struct perf_device;

struct perf_version {
	const char *git_version;
	const char *version;
};

/**
 * type of core: 2D, 3D, VG.
 */
enum perf_core_type {
	PERF_CORE_INVALID = 0,
	PERF_CORE_3D,
	PERF_CORE_2D,
	PERF_CORE_VG,

	PERF_CORE_NUM,
};

/*
 * for systems with multiple cores. Used in perf_read_register()
 * or perf_write_register().
 */
enum perf_core_mpgu_id {
	PERF_MGPU_3D_CORE_0,
	PERF_MGPU_3D_CORE_1,

	PERF_MGPU_3D_CORE_NUM,
};

/**
 * structure to hold driver information. See perf_get_driver_info().
 */
struct perf_driver_info {
	int major, minor;
	int patch, build;
};

/**
 * structure to hold hardware information. See perf_get_hw_info() to retrieve 
 * hw info and perf_free_hw_info() to deallocate memory after you've finished.
 */
struct perf_hw_info {
	uint32_t id;
	uint32_t model;
	uint32_t revision;
	uint32_t product;

	struct perf_hw_info *next;
	struct perf_hw_info *head;
};

/**
 * processes' memory attached to the GPU. See perf_get_client_memory(), as 
 * to retrieve this information and store it in this structure.
 */
struct perf_client_memory {
	uint64_t total;

	uint64_t reserved;
	uint64_t contigous;
	uint64_t _virtual;
	uint64_t non_paged;

	float idle;
};


/**
 * structure holding counter detailed information.
 *
 * See perf_ops::get_counter_info, or perf_get_counter_info().
 */
struct perf_counter_info {
    const char *name;
    const char *desc;
};

/**
 * perf operations that lower-layer library has to implement
 *
 * Not all methods are required to be implemented.
 *
 * Upper-layers have to provide this structure and initialize it
 * with valid functions.
 *
 * With a valid perf_ops structure in place, call perf_init().
 */
struct perf_ops {
	/**
	 * opens a connection with the driver
	 * \param flags: pass certain flags to the driver
	 * \param dev: see perf_init()
	 * \returns 0 in case of success, < 0 otherwise
	 */
	int (*open)(uint32_t flags, struct perf_device *dev);

	/**
	 * uses an already opened connection with the driver using the file descriptor
	 * \param fd: a valid file descriptor
	 * \param flags: pass certain flags to the driver
	 * \param dev: see perf_init()
	 * \returns 0 in case of success, < 0 otherwise
	 */
	int (*open_with_fd)(int fd, uint32_t flags, struct perf_device *dev);

	/**
	 * closes connection to the driver
	 * \param dev: set perf init
	 */
	int (*close)(struct perf_device *dev);

	/**
	 * enables the profiler, enabling the HW to start recording counted values
	 * \param dev: see perf_init()
	 * \returns 0 for success or < 0 otherwise.
	 */
	int (*enable)(struct perf_device *dev);

	/**
	 * enables the profiler but uses the supplied context. Alternative would be
	 * to use perf_context_set() and then call perf_profiler_enable().
	 * \param ctx: an context id
	 * \param dev: see perf_init()
	 * \returns 0 in case of success, < 0 otherwise
	 */
	int (*enable_with_ctx)(uint32_t ctx, struct perf_device *dev);

	/**
	 * disables the profiler.
	 * \param dev: see perf_init()
	 * \returns 0 in case of success, < 0 otherwise
	 */
	int (*disable)(struct perf_device *dev);

	/**
	 * starts the profiler, but unlike perf_profiler_enable() this enables the software
	 * part in the driver. This can be a no-op for upper-layers that do everything in 
	 * perf_profiler_enable()
	 * \param dev: see perf_init()
	 * \returns 0 in case of success, < 0 otherwise
	 */
	int (*start)(struct perf_device *dev);


	/**
	 * stops the profiler this is the reverse of perf_profiler_start()
	 * \param dev: see perf_init()
	 * \returns 0 in case of success, < 0 otherwise
	 */
	int (*stop)(struct perf_device *dev);

	/**
	 * read counter values of type \param type and dump them to \param data
	 * \param type: for VIVANTE one of vivante_profiler_type_counter
	 * \param data: pre-allocated by user, where to dump the counter values
	 * \param dev: see perf_init()
	 * \returns 0 in case of success, < 0 otherwise
	 *
	 * See perf_ops::context_set before using this function.
	 */
	int (*read_counters_3d)(uint32_t type, void *data, struct perf_device *dev);

	/**
	 * read 2D counter values
	 * \param data: where to dump the counter values.
	 * \param dev: see perf_init()
	 * \returns 0 in case of success, < 0 otherwise
	 */
	int (*read_counters_2d)(void *data, struct perf_device *dev);

	/**
	 * reads the register \param reg into \param data
	 * \param core_id: specify one of perf_core_mpgu_id. For single GPUs this is always 0.
	 * \param reg: register address
	 * \param data: where to write the value
	 * \param dev: see perf_init()
	 * \returns 0 in case of success, < 0 otherwise
	 */
	int (*read_register)(uint32_t core_id, uint32_t reg, uint32_t *data, struct perf_device *dev);

	/**
	 * writes the \param data value into \param reg
	 * \param core_id: specify one of perf_core_mpgu_id
	 * \param reg: registers address
	 * \param data: what to write
	 * \param dev: see perf_init()
	 * \returns 0 in case of success, < 0 otherwise
	 */
	int (*write_register)(uint32_t core_id, uint32_t reg, uint32_t data, struct perf_device *dev);

	/**
	 * get the number of available counters 
	 * \param type: for VIVANTE see vivante_profiler_type_counter
	 * \returns number of counters in case of success, or 0 otherwise.
	 */
	uint32_t (*get_num_counters)(uint32_t type);

	/**
	 * Retrieve detailed information about counter \param id of type \param type
	 * \param type: one of vivante_profiler_type_counter
	 * \param id: a unsigned integer
	 * \returns a pointer to struct perf_counter_info or NULL otherwise.
	 */
	struct perf_counter_info *(*get_counter_info)(uint32_t type, uint32_t id);

	/**
	 * Reset counter values based on \param val
	 * \param bool: if TRUE, resets the counter values upon reading them
	 * again.
	 * \param dev: a pointer to a structure perf_device
	 * \returns 0 in case of scucess, < 0 otherwise
	 */
	int (*reset_counters)(bool val, struct perf_device *dev);

	/**
	 * Determine if proiler is active in the driver. Might be a no-op for
	 * non-VSI driver.
	 * \param state: A integer used to check what kind of errors we might
	 * receive, in case the return value is < 0.
	 * \returns 0 in case of scucess, < 0 otherwise
	 */
	int (*profiler_is_available)(int *state);

	/**
	 * Retrieve current contexts id of the running process.
	 * \param dev: a pointer to structure perf_device
	 * \param ctx_no: An unsigned int pointer to store the nr of 
	 * contexts found. Useful to know the size when iterating over the
	 * array returned.
	 * \param pid: an unsigned integer of the pid.
	 * \returns a statically allocated array with contexts or NULL
	 * otherwise. There's no need to de-allocate in case of success.
	 *
	 * This method is useful when trying to find the contexts associated
	 * with the running process.
	 */
	int (*get_contexts)(uint32_t **ctxs_no, uint32_t *ctx_no, uint32_t pid, struct perf_device *dev);

	/**
	 * Sets context id to value specified in \param ctx
	 * \param ctx: An unsigned integer.
	 * \param dev: a pointer to struct perf_device. See perf_init().
	 * \returns 0 in case the context was successfully changed or 
	 * < 0 otherwise.
	 *
	 * This method could be a no-op for non-VSI driver. Otherwise this
	 * method has to be called if perf_profiler_enable_with_ctx() has *not*
	 * been called. This is required before invoking
	 * perf_read_counters_3d().
	 */
	int (*context_set)(uint32_t ctx, struct perf_device *dev);

	/**
	 * retrive information about driver 
	 * \param drv_info: a pointer to struct perf_driver_info to store driver
	 * information.
	 * \param dev:  a pointer to struct perf_device
	 * \returns 0 in cae of success, < 0 otherwise.
	 */
	int (*get_driver_info)(struct perf_driver_info *drv_info, struct perf_device *dev);

	/**
	 * retrive information about hw_info. Users should call perf_free_hw_info()
	 * after getting/using this information.
	 * \param hw_info: Where to store hardware information about the GPU.
	 * \param dev: a pointer to struct perf_device
	 * \returns 0 in case of success, < 0 otherwise
	 */
	int (*get_hw_info)(struct perf_hw_info *hw_info, struct perf_device *dev);

	/**
	 * call this method after you've called perf_get_hw_info() to release
	 * memory.
	 * \param hw_info: a pointer to structure perf_hw_info
	 * \returns 0 in case of success, < 0 otherwise
	 */
	int (*free_hw_info)(struct perf_hw_info *hw_info);

	/**
	 * retrieve memory information specified in \param pid 
	 * \param c_mem: a pointer to a struct perf_client_memory. Where to dump
	 * that information.
	 * \param pid: pid of the process we want to retrieve memory info
	 * \param dev: see perf_init()
	 * \returns 0 in case of success and c_mem is populated, or -1 otherwise
	 */
	int (*get_client_memory)(struct perf_client_memory *c_mem, int pid, struct perf_device *dev);

	/**
	 * Get type of core, based on \param core_id. This function chould be a
	 * no-op to other users than VSI, as this VSI encodes its core_id as
	 * unsigned integers.
	 * \param core_id: An unsigned integer. Could be retrieved using
	 * perf_get_hw_info() and by looking at perf_hw_info::id.
	 * \param dev: a pointer to struct perf_device
	 * \returns a perf_core_type or PERF_CORE_INVALID otherwise.
	 */
	enum perf_core_type (*get_core_type)(uint32_t core_id, struct perf_device *dev);

	/**
	 * retrieve the last error gpuperfcnt gave
	 * \param dev see perf_init()
	 * \return a pointer to a const char 
	 */
	const char *(*get_last_error)(struct perf_device *dev);
};


/**
 * holds the private implementation and the perf operations
 */
struct perf_device {
	void *_private;		/** private implementation of the underlying driver */
	struct perf_ops *ops;	/** a pointer to struct perf_ops */
};


/**
 * creates a perf_device
 * \param perf_ops specify which operations should have this perf device
 * \return a pointer to an allocated struct perf_device or NULL otherwise
 *
 * Upon calling this function, other perf_ functions will directly use 
 * the callbacks set in the perf_ops structure, so make sure you initialize
 * perf_ops before.
 *
 * perf_exit() shall be called to close the connection with the driver and
 * de-allocate resources.
 */
struct perf_device *
perf_init(struct perf_ops *perf_ops);


/**
 * opens a communication channel with the driver
 * \param args additional arguments to supply to underlying driver
 * \param dev a pointer to a struct perf_device created with perf_init()
 * \return 0 in case of success, -1 otherwise.
 */
int
perf_open(uint32_t args, struct perf_device *dev);

/**
 * opens a communication channel with the driver by supplying a file-descriptor
 * specified in \param fd
 * \param fd the file description that talks with the driver
 * \param args additional arguments to pass to the driver
 * \param dev a pointer to struct perf_device
 * \return 0 in case of success, -1 otherwise.
 *
 * This method can be used in case the driver doesn't allow same process to open
 * multiple fds.
 */
int
perf_open_with_fd(int fd, uint32_t args, struct perf_device *dev);

/**
 * Enables profiling.
 * \param dev a pointer to a structure perf_device
 * \return 0 in case of success, -1 otherwise.
 */
int
perf_profiler_enable(struct perf_device *dev);


/**
 * enables profiling with a context provided in \param ctx
 * \param ctx an  unsigned integer
 * \param dev a pointer to a structure perf_device
 * \return 0 in case of success, -1 otherwise.
 *
 * This method equivalent to perf_profiler_enable() and
 * perf_context_set().
 */
int
perf_profiler_enable_with_ctx(uint32_t ctx, struct perf_device *dev);

/**
 * disables profiling
 * \param dev a pointer to a structure perf_device
 * \return 0 in case of success, -1 otherwise.
 */
int
perf_profiler_disable(struct perf_device *dev);

/**
 * starts profiling. This method is used in cases where additional 
 * commands have to be issued to the driver.
 * \param dev a pointer to a structure perf_device
 * \return 0 in case of success, -1 otherwise.
 *
 * This method should be called before calling perf_read_counters_3d().
 */
int
perf_profiler_start(struct perf_device *dev);

/**
 * stops the profiler.
 * \param dev a pointer to a structure perf_device
 * \return 0 in case of success, -1 otherwise.
 */
int
perf_profiler_stop(struct perf_device *dev);

/**
 * gets the number of available hardware-counters
 * \param type one of vivante_profiler_type_counter
 * \param dev see perf_init()
 * \return 0 in case of success, -1 otherwise.
 */
uint32_t
perf_get_num_counters(uint32_t type, struct perf_device *dev);

/**
 * read in the values of the HW counters
 * \param type one of vivante_profiler_type_counter
 * \param data where to dump the data. Must be allocated by the caller.
 * \param dev see perf_init()
 * \return 0 in case of success, -1 otherwise.
 *
 * Note that, before this function is called, the profiler must be enabled
 * and started. Additionally a context must be set.
 *
 * For instance
 *
 * // enable and set a context
 * perf_profiler_enable(dev);
 * perf_context_set(ctx, dev);
 *
 * // start the profiler 
 * perf_profiler_start(dev);
 *
 * //now we can safely call perf_read_counters_3d
 * perf_read_counters_3d(type, data, dev);
 */
int
perf_read_counters_3d(uint32_t type, void *data, struct perf_device *dev);

/**
 * read int counter values. 2D version.
 * \param data where to dump the data. Must be allocated by the user.
 * \param dev see perf_init()
 * \return 0 in case of success, -1 otherwise.
 */
int
perf_read_counters_2d(void *data, struct perf_device *dev);


/**
 * reads register \param reg into \param data
 * \param cid specify the core_id, one of perf_core_mpgu_id
 * \param reg the address of the register
 * \param data where to dump the data
 * \param dev a pointer to structure perf_device
 * \return 0 in case of success, -1 otherwise.
 */
int
perf_read_register(uint32_t cid, uint32_t reg, uint32_t *data, struct perf_device *dev);

/**
 * write register \param reg from \param data
 * \param cid specify the core_id, one of perf_core_mpgu_id
 * \param reg the address of the register
 * \param data where to dump the data
 * \param dev a pointer to structure perf_device
 * \return 0 in case of success, -1 otherwise.
 */
int
perf_write_register(uint32_t cid, uint32_t reg, uint32_t data, struct perf_device *dev);

/**
 * resets the counter values after they have been read.
 * \param val in case of TRUE, next time perf_read_counters_3d() will be called 
 * would reset the counter values, after receiving the counter values.
 * \param dev see perf_init()
 * \return 0 in case of success, -1 otherwise.
 *
 * Note that this method is a way of hinting the driver to reset the values they
 * have been read.
 */
int
perf_reset_counters(bool val, struct perf_device *dev);


/**
 * get counter detailed information
 * \param type one of vivante_profiler_type_counter
 * \param id the id (unsigned int)
 * \param dev see perf_init
 * \return returns a pointer to a structure of perf_counter_info
 * type or NULL if id is not found.
 *
 * This method should be paired with perf_get_num_counters().
 *
 * For instance:
 *
 * for (size_t i = 0; i < perf_get_num_counters(type, dev); i++)
 *	struct perf_counter_info * = perf_get_counter_info(type, i, dev);
 */
struct perf_counter_info *
perf_get_counter_info(uint32_t type, uint32_t id,  struct perf_device *dev);

/**
 * check for profiler support
 * \param state: 0 in case profiler is enabled, -1 if gpu profiler is not enabled or
 * -2 in case power management is not disabled
 * \param dev see perf_init()
 * \return 0 in case profiler is enabled, -1 otherwise.
 */
int
perf_check_profiler(int *state, struct perf_device *dev);

/**
 * returns context of the application specified in \param pid
 * \param ctxs a statically allocated array with the contexts
 * \param ctx_no the nr of contexts retrieved
 * \param pid the pid that you want to retrieve the contexts.
 * \param dev  a pointer to struct perf_device
 * \return 0 in case of success, -1 otherwise.
 */
int
perf_get_contexts(uint32_t **ctxs, uint32_t *ctx_no, uint32_t pid, struct perf_device *dev);

/** 
 * Set the context to \param ctx This functions needs to be used
 * in tandem before/with perf_read_counters_3d(). 
 *
 * Use perf_get_current_contexts() or perf_get_contexts() go retrieve the
 * contexts, or use find the context id by using debugfs or other means.
 *
 * \param ctx an unsigned int (a valid context id)
 * \param dev a pointer to struct perf_device
 * \return 0 in case of success, -1 otherwise.
 */
int
perf_context_set(uint32_t ctx, struct perf_device *dev);

/**
 * retrieve driver information and store it in \param drv_info
 * \param drv_info a pointer where to store the driver information
 * \param dev see perf_init()
 * \return 0 in case of success, -1 otherwise.
 */
int
perf_get_driver_info(struct perf_driver_info *drv_info, struct perf_device *dev);

/**
 * Retrieve hw information 
 * \param hw_info a pointer where to store hardware information about the GPU
 * \param dev see perf_init
 * \return 0 in case of success, -1 otherwise.
 *
 * This method allocates memory so call perf_free_hw_info() after you finished
 * with the info in structure perf_hw_info.
 */
int
perf_get_hw_info(struct perf_hw_info *hw_info, struct perf_device *dev);

/**
 * Frees mermory allocated by perf_get_hw_info()
 * \param hw_info pointer to struct perf_hw_info
 * \param dev see perf_init
 * \return 0 in case of success, -1 otherwise.
 */
int
perf_free_hw_info(struct perf_hw_info *hw_info, struct perf_device *dev);

/**
 * retrieve client memory and store it in pointer to \param cmem
 * \param cmem a pointer to structure perf_client_memory to store the
 * information
 * \param pid a unsigned integer, specify process pid
 * \param dev a pointer to struct perf_device
 * \return 0 in case of success, -1 otherwise.
 */
int
perf_get_client_memory(struct perf_client_memory *cmem, int pid, struct perf_device *dev);


/**
 * retrieve the type of core specified by \param id
 * \param id an unsigned int
 * \param dev a pointer to a struct perf_device
 * \return one of enum perf_core_type or PERF_CORE_INVALID otherwise
 *
 * This function is useful in converting between VSI specific encoding
 * of cores to a more human-like presentation.
 */
enum perf_core_type 
perf_get_core_type(uint32_t id, struct perf_device *dev);

/**
 * retrieve the last error the library gave
 * \param dev a pointer to 
 * \return a pointer to a char or NULL in case no error was previously stored.
 *
 * This method can be called when a function returns < 0 but you want to
 * provide a human-like description of the error.
 *
 */
const char *
perf_get_last_error(struct perf_device *dev);

/**
 * retrieve library version
 */
struct perf_version
perf_get_library_version(void);

/**
 * Closes and frees the perf_device created with perf_init()
 * \param dev see perf_init()
 *
 * This functions should be called after you've finished with the
 * library calls.
 */
void
perf_exit(struct perf_device *dev);

#ifdef __cplusplus
}
#endif

#endif /* __GPUPERFCNT_H */
